declare global {
  interface Window {
    JikeOpenHybrid: {
      nativeDispatch?: any
      webDispatch?: any
      callbacks?: { [key: string]: Function }
    }
  }
  const SDK_VERSION: string
}

import * as uuid from 'uuid'
import 'reflect-metadata'
import { availableVersion } from './utils'
import {
  ActionSchema,
  JikeOpenJsSDKConfiguration,
  ActionCallback,
  ActionMetadata,
  UserInfo,
  APISchema,
} from './interfaces'

const UA = () => window.navigator.userAgent

window.JikeOpenHybrid = window.JikeOpenHybrid || {}
window.JikeOpenHybrid.callbacks = {}
window.JikeOpenHybrid.webDispatch = (action: ActionSchema<any>) => {
  if (window.JikeOpenHybrid.callbacks[action.type]) {
    window.JikeOpenHybrid.callbacks[action.type](
      action.error,
      action.payload,
      (action.callback || {}).context,
    )
  }
}

const JikeOpenHybrid = window.JikeOpenHybrid

/**
 * jike open platform javascript sdk.
 *
 * @export
 * @class JikeOpenJsSDK
 */
export class JikeOpenJsSDK {
  private _configuration: JikeOpenJsSDKConfiguration

  public get configuration(): JikeOpenJsSDKConfiguration {
    return this._configuration
  }
  public set configuration(value: JikeOpenJsSDKConfiguration) {
    this._configuration = value
  }

  constructor(value: string | JikeOpenJsSDKConfiguration) {
    if (typeof value === 'string') {
      this.configuration = { openAppId: value }
    } else {
      this.configuration = value
    }
  }

  /**
   * Check whether JikeOpenHybrid Object is loaded.
   *
   * @todo Reject condition?
   * @param {Function} [cb] Callback function.
   * @memberof JikeOpenJsSDK
   */
  public ready() {
    return new Promise(resolve => {
      const intervalId = setInterval(() => {
        if (JikeOpenHybrid.nativeDispatch) {
          clearInterval(intervalId)
          resolve()
        }
      }, 10)
    })
  }

  /**
   * Get user info.
   * @deprecated
   * @returns User info (just for open platform part).
   * @memberof JikeOpenJsSDK
   */
  @availableVersion('4.10.0')
  public getUserInfo_deprecated() {
    const callback: ActionCallback = {
      actionType: 'getUserInfoCallback',
    }
    return this.callNative<{}, UserInfo>('jkopen_user_info', {}, callback).then(({ payload }) => {
      return payload
    })
  }

  /**
   * Get user info.
   *
   * @returns
   * @memberof JikeOpenJsSDK
   */
  @availableVersion('4.15.0')
  public async getUserInfo() {
    const callback: ActionCallback = {
      actionType: 'getUserInfoCallback',
    }
    const { data } = await this.callAPI<{ data: UserInfo }>(
      {
        resources: 'users',
        action: 'me',
      },
      callback,
    )
    return data
  }

  @availableVersion('4.15.0')
  public async getMessages(topicId: string) {
    const callback: ActionCallback = {
      actionType: 'getMessagesCallback',
    }
    const { data } = await this.callAPI<{ data: { message: Array<any>, count: number } }>(
      {
        resources: 'messages',
        action: 'list',
        params: { topic: topicId },
      },
      callback,
    )
    return data
  }

  /**
   *
   * @private
   * @template T
   * @param {APISchema} apiSchema
   * @param {ActionCallback} [callback]
   * @returns
   * @memberof JikeOpenJsSDK
   */
  @availableVersion('4.15.0')
  private async callAPI<T>(apiSchema: APISchema, callback?: ActionCallback) {
    const {
      payload: { result },
    } = await this.callNative<APISchema, { result: T }>(
      'jkopen_api',
      {
        ...apiSchema,
        method: apiSchema.method || 'GET',
        params: apiSchema.params || {},
      },
      callback,
    )
    return result
  }

  /**
   * Check whether runtime is in Jike App.
   *
   * @returns {boolean}
   * @memberof JikeOpenJsSDK
   */
  public isInJikeApp(): boolean {
    return !!UA().match(/jike/i)
  }

  /**
   * Get Jike App version.
   *
   * @returns {string} App version
   * @memberof JikeOpenJsSDK
   */
  public appVersion(): string {
    const matched = UA().match(/jike\/([0-9]*).([0-9]*).([0-9]*)/i)
    return matched ? matched[0].split('/')[1] : ''
  }

  /**
   * Compare input target version with current runtime version.
   *
   * @param {string} version Target version
   * @returns {number} 0: equal, 1:  higher, -1: lower
   * @memberof JikeOpenJsSDK
   */
  public compareVersion(version: string): number {
    const appVersion = this.appVersion()
    const appVersionList = appVersion.split('.').map(s => parseInt(s) || 0)
    const versionList = version.split('.').map(s => parseInt(s) || 0)
    for (let i = 0; i < 3; i++) {
      if (appVersionList[i] > versionList[i]) {
        return 1
      } else if (appVersionList[i] < versionList[i]) {
        return -1
      }
    }
    return 0
  }

  /**
   * Check whether input function can be use in runtime version.
   *
   * @todo whether needisInJikeApp
   * @param {string} funcName Function name.
   * @returns {boolean}
   * @memberof JikeOpenJsSDK
   */
  public checkAvailable(funcName: string): boolean {
    return (
      this.isInJikeApp() &&
      this.compareVersion(Reflect.getMetadata('iOSVersion', this[funcName])) > -1 &&
      this.compareVersion(Reflect.getMetadata('androidVersion', this[funcName])) > -1
    )
  }

  /**
   * Call native method
   *
   * @private
   * @template T
   * @template R
   * @param {string} type
   * @param {T} [payload]
   * @param {ActionCallback} [callback]
   * @param {ActionMetadata} [meta]
   * @returns
   * @memberof JikeOpenJsSDK
   */
  private callNative<T, R>(
    type: string,
    payload?: T,
    callback?: ActionCallback,
    meta?: ActionMetadata,
  ) {
    const callbackId = callback.actionType || uuid()
    const action: ActionSchema<T> = {
      type,
      payload,
      meta: {
        ...meta,
        openAppId: this.configuration.openAppId,
        sdkVersion: SDK_VERSION,
      },
      callback: callback || {
        actionType: callbackId,
        context: {},
      },
    }
    const promise = new Promise<{ payload: R; context: any }>((resolve, reject) => {
      window.JikeOpenHybrid.callbacks[callbackId] = (error, pl, context) => {
        delete window.JikeOpenHybrid.callbacks[callbackId]
        if (error) {
          reject({ payload: pl, context })
        } else {
          resolve({ payload: pl, context })
        }
      }
    })
    JikeOpenHybrid.nativeDispatch(JSON.stringify(action))
    return promise
  }

  /**
   * Call native method without result
   *
   * @private
   * @template T
   * @param {string} type
   * @param {T} [payload]
   * @param {ActionCallback} [callback]
   * @param {ActionMetadata} [meta]
   * @memberof JikeOpenJsSDK
   */
  private callNativeWithoutResult<T>(
    type: string,
    payload?: T,
    callback?: ActionCallback,
    meta?: ActionMetadata,
  ) {
    const action: ActionSchema<T> = {
      type,
      payload,
      meta: {
        ...meta,
        openAppId: this.configuration.openAppId,
        sdkVersion: SDK_VERSION,
      },
      callback,
    }
    JikeOpenHybrid.nativeDispatch(JSON.stringify(action))
  }
}
